#include "boost/numeric/ublas/vector_proxy.hpp"
#include "boost/numeric/ublas/matrix_proxy.hpp"
#include "hpp/hik/solver-basic.hh"

#include <time.h>
#include <sys/time.h>

using namespace boost::numeric::ublas;

ChppHikSolverBasic::ChppHikSolverBasic(unsigned int inNparam)
{
    attNParam = inNparam;

    attSlackSolver = new ChppHikSlackSolver(inNparam);
    attEqConstraint = new ChppHikLinearSystem(inNparam);
    attIneqConstraint = new ChppHikLinearSystem(inNparam);
    attEqObjective = new ChppHikLinearSystem(inNparam);
    attIneqObjective = new ChppHikLinearSystem(inNparam);

    attThresh = 1e-5;

    attMask = scalar_vector<double>(inNparam,1);

    attSolution.resize(attNParam);
    attSolution.clear();

}

ChppHikSolverBasic::~ChppHikSolverBasic()
{
    delete attSlackSolver;
    delete attEqConstraint;
    delete attEqObjective;
    delete attIneqConstraint;
    delete attIneqObjective;
}

void ChppHikSolverBasic::upperBound(unsigned int i_rank, double i_bound)
{
    attSlackSolver->upperBound( i_rank, i_bound );
}

void ChppHikSolverBasic::lowerBound(unsigned int i_rank, double i_bound)
{
    attSlackSolver->lowerBound( i_rank, i_bound );
}

void ChppHikSolverBasic::defaultUpperBound(unsigned int i_rank, double i_bound)
{
    attSlackSolver->defaultUpperBound( i_rank, i_bound );
}

void ChppHikSolverBasic::defaultLowerBound(unsigned int i_rank, double i_bound)
{
    attSlackSolver->defaultLowerBound( i_rank, i_bound );
}

void ChppHikSolverBasic::clearBounds()
{
    attSlackSolver->clearBounds();
}

bool ChppHikSolverBasic::setMask(const vectorN& inMask)
{
    if (inMask.size() != attMask.size())
        return false;

    attMask = inMask;
    return true;
}

void ChppHikSolverBasic::initFeasibleSet()
{
    /*
    struct timeval *Tps, *Tpf;
    struct timezone *Tzp;

    
    Tps = ( struct timeval* ) malloc ( sizeof ( struct timeval ) );
    Tpf = ( struct timeval* ) malloc ( sizeof ( struct timeval ) );
    Tzp = 0;
    gettimeofday ( Tps, Tzp );
    */
    attEqConstraint->clear();
    attIneqConstraint->clear();

    for(unsigned int i=0;i<attEqConstraints.size();i++)
        attEqConstraint->expand(attEqConstraints[i]->jacobian(), attEqConstraints[i]->value());

    for(unsigned int i=0;i<attIneqConstraints.size();i++)
        attIneqConstraint->expand(attIneqConstraints[i]->jacobian(), attIneqConstraints[i]->value());

    attEqConstraint->disableColumns( attMask );
    attIneqConstraint->disableColumns( attMask );
    /*
    gettimeofday ( Tpf, Tzp );
    printf ( "initFeasibleSet phase costs: %ld us\n", ( Tpf->tv_sec-Tps->tv_sec ) *1000000 + ( Tpf->tv_usec-Tps->tv_usec ) );
    
    free ( Tps );
    free ( Tpf );*/
}

void ChppHikSolverBasic::clearTasks()
{
    attTasks.clear();
}

bool ChppHikSolverBasic::addTask(CjrlLinearConstraint& inTask, bool inIsEqualityTask)
{
    if (inTask.jacobian().size2() != attNParam)
    {
        std::cout << "ChppHikSolverBasic::addTask failed" <<std::endl;
        return false;
    }
    attTasks.push_back(std::make_pair(&inTask,inIsEqualityTask));
    return true;
}

int ChppHikSolverBasic::solve(const vectorN& inInitialGuess, double inDampingFactor)
{

    if (inInitialGuess.size()!=attNParam)
    {
        std::cout << "ChppHikSolverBasic::solve failed. First argument of incorrect size."<<std::endl;
        return -1000;
    }
    attSolution = inInitialGuess;

    attEqObjective->clear();
    attIneqObjective->clear();

    for (unsigned int i=0; i<attTasks.size();i++)
        if (attTasks[i].second)
            attEqObjective->expand(attTasks[i].first->jacobian(), attTasks[i].first->value());
        else
            attIneqObjective->expand(attTasks[i].first->jacobian(), attTasks[i].first->value());

    attEqObjective->disableColumns( attMask );
    attIneqObjective->disableColumns( attMask );


    attSlackSolver->clearTasks();
    attSlackSolver->clearConstraints();
    attSlackSolver->addTasks(attIneqObjective, false);
    attSlackSolver->addTasks(attEqObjective, true );
    attSlackSolver->addConstraints(attIneqConstraint, false);
    attSlackSolver->addConstraints(attEqConstraint, true );

    
    
    
    
    
    bool solved = attSlackSolver->solve(attSolution,attSlacks, inDampingFactor);
    
    
    return solved;
}

const vectorN& ChppHikSolverBasic::solution()
{
    return attSolution;
}


void ChppHikSolverBasic::updateFeasibleSet()
{
    unsigned int curStart = 0, i;

    for (i=0; i<attTasks.size();i++)
    {
        if (attTasks[i].second == false)//inequality
        {
            CjrlLinearConstraint& inTask = *(attTasks[i].first);
            tempb = subrange(attSlacks,curStart,curStart+inTask.dimension());
            std::vector<unsigned int> unsatsf,satsf;
            for (unsigned int i=0; i<tempb.size();i++)
            {
                if (tempb(i) < attThresh)
                    satsf.push_back(i);
                else
                    unsatsf.push_back(i);
            }

            attEqConstraint->expand(inTask.jacobian(),tempb+inTask.value(), &unsatsf);
            attIneqConstraint->expand(inTask.jacobian(), inTask.value(),&satsf);
            curStart += inTask.dimension();
        }
    }

    for (i=0; i<attTasks.size();i++)
    {
        if (attTasks[i].second == true)//equality
        {
            CjrlLinearConstraint& inTask = *(attTasks[i].first);
            tempb = prod(inTask.jacobian(), attSolution);
            attEqConstraint->expand(inTask.jacobian(), tempb);
        }
    }

    attEqConstraint->disableColumns( attMask );
    attIneqConstraint->disableColumns( attMask );
}

void ChppHikSolverBasic::removePreConstraints()
{
    attEqConstraints.clear();
    attIneqConstraints.clear();
}

void ChppHikSolverBasic::addEqPreConstraint( CjrlLinearConstraint& inConstraint)
{
    attEqConstraints.push_back(&inConstraint);
}

void ChppHikSolverBasic::addIneqPreConstraint( CjrlLinearConstraint& inConstraint)
{
    attIneqConstraints.push_back(&inConstraint);
}

CjrlLinearConstraint& ChppHikSolverBasic::currentFeasibleSetEquality()
{
    return *attEqConstraint;
}

CjrlLinearConstraint& ChppHikSolverBasic::currentFeasibleSetInequality()
{
    return *attIneqConstraint;
}
